@@ -3,11 +3,11 @@ try:
 except ImportError:
     pass  # not jupyter.
 
-import json, re, os
+import json, re, os, pickle
 from warnings import warn
 from typing import Any
 from datetime import datetime
-from typing import Optional, Dict
+from typing import Optional, Dict, List
 
 from .enums import Privacy, Location
 
@@ -154,7 +154,11 @@ class MikePage(TableMixin):
     # ==== Columns =====================================================================================================
 
     def check_columns(self):
-        if self.columns_viewport + self._columns_text != 12:
+        """
+        Verify that columns_viewport and columns_text are 12.
+        See bootstrap docs for details of why a full row is 12.
+        """
+        if self._columns_viewport + self._columns_text != 12:
             raise ValueError('text and viewport are not 12.')
 
     def _get_columns_viewport(self):
@@ -213,7 +217,12 @@ class MikePage(TableMixin):
             setattr(self, k, v)
         return self
 
-    def dump(self):
+    def dumps(self):
+        """
+        For proper saving see ``to_pickle`` or ``save``.
+        convert the object as dictionary.
+        :return:
+        """
         data = self.__dict__.copy()
         data['page'] = self.page
         data['proteinJSON'] = json.dumps(self.proteins)
@@ -265,14 +274,34 @@ class MikePage(TableMixin):
         """
         self.refresh_image = True
 
-    def save(self, name:str):
+    def to_pickle(self, filename: str):
         """
-        Save the description, pdbs and JS for easier editing
+        writes to pickle
 
-        :param name:  filename (path okay) with no extension.
+        :param filename:  filename (path okay) with extension.
         :return:
         """
-        with open(name+'.md', 'w') as w:
+        pickle.dump(self.dumps(), open(filename,'wb'))
+
+    def from_pickle(self, filename: str):
+        data = pickle.load(open(filename,'wb'))
+        self.page = data['page']
+        self.proteins = json.loads(data['proteinJSON'])
+        self.pdbs = {k: v for k, v in json.loads(data['pdb'])}
+        self.location_viewport.name = data['location_viewport']
+        self.public.name = data['public']
+
+    def save(self, name: str):
+        """
+        Save the description, pdbs and JS for easier editing.
+        It's inverse ``load`` loads them.
+        The purpose is to edit the file as opposed to a wonky notebook cell.
+
+        ``save`` is a legacy keyword in the page data, but is no longer used.
+        :param name:
+        :return:
+        """
+        with open(name + '.md', 'w') as w:
             w.write(self.description)
         with open(name + '.js', 'w') as w:
             w.write(self.loadfun)
@@ -280,14 +309,14 @@ class MikePage(TableMixin):
             with open(f'{name}-{pdbname}.pdbs', 'w') as w:
                 w.write(self.pdbs[pdbname])
 
-    def load(self, name:str):
+    def load(self, name: str):
         """
-        Loads what was saved with ``.save``
+        Loads what was saved with ``.saved``
 
         :param name:  filename (path okay) with no extension.
         :return:
         """
-        with open(name+'.md', 'r') as r:
+        with open(name + '.md', 'r') as r:
             self.description = r.read()
         with open(name + '.js', 'r') as r:
             self.loadfun = r.read()
@@ -297,11 +326,11 @@ class MikePage(TableMixin):
 
     # ======== Parent ==================================================================================================
 
-    def rename_protein_variable(self, 
-                       index: Optional[int] = None,
-                       name: Optional[str] = None,
-                       value: Optional[str] = None,
-                       newname: Optional[str] = None):
+    def rename_protein_variable(self,
+                                index: Optional[int] = None,
+                                name: Optional[str] = None,
+                                value: Optional[str] = None,
+                                newname: Optional[str] = None):
         """
         Overloaded method. accepts one of the three parameters:
         
@@ -315,8 +344,8 @@ class MikePage(TableMixin):
         oldname = protein['value']
         if newname and re.search('[^\w\_]', newname) is not None:
             raise ValueError(f'{newname} is not legal')
-        elif newname and 'isVariable' in protein and protein['isVariable'] in (True, 'true'): # rename
-            protein['value'] = re.replace('[^\w\_]','', newname)
+        elif newname and 'isVariable' in protein and protein['isVariable'] in (True, 'true'):  # rename
+            protein['value'] = re.sub('[^\w\_]', '', newname)
             self.pdbs[newname] = self.pdbs[oldname]
             del self.pdbs[oldname]
         elif newname:
@@ -325,11 +354,11 @@ class MikePage(TableMixin):
             protein['value'] = self.pdbs[oldname]
             del protein['isVariable']
             del self.pdbs[oldname]
-    
+
     def get_protein(self,
                     index: Optional[int] = None,
-                       name: Optional[str] = None,
-                       value: Optional[str] = None) -> Dict:
+                    name: Optional[str] = None,
+                    value: Optional[str] = None) -> Dict:
         """
         Overloaded method. accepts one of the three parameters:
         
@@ -352,16 +381,17 @@ class MikePage(TableMixin):
         else:
             raise ValueError('Impossible.')
 
-    def make_github_entry(self, username: str, repo: str, path: str) -> Dict:
+    def make_github_entry(self, username: str, repo: str, path: str, branch: str='main') -> Dict:
         """
         make a protein entry.
 
         :param username: Github username
         :param repo: repository
         :param path: path within repo
+        :param branch: main or master etc.
         :return:
         """
-        url = f'https://raw.githubusercontent.com/{username}/{repo}/master/{path}'
+        url = f'https://raw.githubusercontent.com/{username}/{repo}/{branch}/{path}'
         name = re.sub('[^\w_]', '', os.path.splitext(os.path.split(path)[1])[0])
         return {'type': 'url', 'value': url, 'name': name}
 
@@ -377,6 +407,86 @@ class MikePage(TableMixin):
         self.proteins.append(self.make_github_entry(username, repo, path))
         return len(self.proteins) - 1
 
+    def append_pdbfile(self, filename: str, varname: Optional[str] = None, chain_definitions: List[dict] = []) -> int:
+        """
+        make and add a protein entry.
+
+        :param filename: the local filename to add.
+        :param varname: what should the pdbblock be known as?
+        :param chain_definitions: list of dict defining who each chain is. Needed solely for Uniprot modal and chain name in builder
+        :return:
+        """
+        pdbblock = open(filename).read()
+        if varname is None:
+            varname = re.sub(r'[^\w_]', '', os.path.splitext(filename)[0])
+            varname = re.sub(r'^\d+', '', varname)
+            print(f'Variable name is {varname}')
+        self.append_pdbblock(pdbblock, varname, chain_definitions)
+        return len(self.proteins) - 1
+
+    def append_pdbblock(self, pdbblock: str, varname: str, chain_definitions: List[dict] = []) -> int:
+        """
+        make and add a protein entry.
+
+        :param pdbblock: the pdbblock to add.
+        :param varname: what should the pdbblock be known as?
+        :param chain_definitions: list of dict defining who each chain is. Needed solely for Uniprot modal and chain name in builder
+        :return:
+        """
+        assert varname not in self.pdbs, f'Varname {varname} already in use.'
+        self.proteins.append({'type': 'data',
+                              'value': varname,
+                              'isVariable': 'true',
+                              'chain_definitions': chain_definitions,
+                              'history': {'code': '', 'changes': ''}})
+        self.pdbs[varname] = pdbblock
+        return len(self.proteins) - 1
+
+    def remove_protein(self, index):
+        if self.proteins[index]['type'] == 'data':
+            del self.pdbs[self.proteins[index]['value']]
+        d = self.proteins[index]
+        del self.proteins[index]
+        return d
+
+    def add_publication(self,
+                        url: Optional[str] = None,
+                        authors: Optional[str] = 'TBA',
+                        year: Optional[int] = None,
+                        title: Optional[str] = 'TBA',
+                        journal: Optional[str] = 'manuscript in preparation',
+                        issue: Optional[str] = 'NA',
+                        doi: Optional[str] = None):
+        """
+        This is an admin only operation due to the security issue (dodgy links in the gallery).
+        Publication is not data from the page, it's an actual table
+
+        :param uuid:
+        :param url:
+        :param authors:
+        :param year:
+        :param title:
+        :param journal:
+        :param issue:
+        :return:
+        """
+        if year is None:
+            year = datetime.year
+        if url is not None:
+            pass
+        elif doi is None:
+            url = '#'
+        else:
+            url = f'https://dx.doi.org/{doi}'
+        return self.parent.post_json('set', {'item': 'publication',
+                                             'identifier': self.page,
+                                             'url': url,
+                                             'title': title,
+                                             'year': year,
+                                             'authors': authors,
+                                             'journal': journal,
+                                             'issue': issue})
+
     # ======== Parent ==================================================================================================
 
     def retrieve(self):
